@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2009-2018 Vadim Mikhailov
+ * Copyright (c) 2009-2019 Vadim Mikhailov
  *
  * Utility to turn USB port power on/off
  * for USB hubs that support per-port power switching.
@@ -29,16 +29,20 @@
 #include <unistd.h>
 #endif
 
-#if defined(__FreeBSD__) || defined(_WIN32)
+#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(_WIN32)
 #include <libusb.h>
 #else
 #include <libusb-1.0/libusb.h>
 #endif
 
-#if defined(__APPLE__) /* snprintf is not available in pure C mode */
+#if defined(__APPLE__) || defined(__FreeBSD__) /* snprintf is not available in pure C mode */
 int snprintf(char * __restrict __str, size_t __size, const char * __restrict __format, ...) __printflike(3, 4);
 #endif
 
+#if !defined(LIBUSB_API_VERSION) || (LIBUSB_API_VERSION <= 0x01000103)
+#define LIBUSB_DT_SUPERSPEED_HUB 0x2a
+#endif
+
 #if _POSIX_C_SOURCE >= 199309L
 #include <time.h>   /* for nanosleep */
 #endif
@@ -59,14 +63,8 @@ void sleep_ms(int milliseconds)
 #endif
 }
 
-/* Max number of hub ports supported.
- * This is somewhat artifically limited by "-p" option parser.
- * If "-p" parser is improved, we can support up to 32 ports.
- * However, biggest number of ports on smart hub I've seen was 8.
- * I've also observed onboard USB hub with whopping 14 ports,
- * but that hub did not support per-port power switching.
- */
-#define MAX_HUB_PORTS            9
+/* Max number of hub ports supported */
+#define MAX_HUB_PORTS            14
 #define ALL_HUB_PORTS            ((1 << MAX_HUB_PORTS) - 1) /* bitmask */
 
 #define USB_CTRL_GET_TIMEOUT     5000
@@ -180,15 +178,24 @@ struct usb_port_status {
 /* List of all USB devices enumerated by libusb */
 static struct libusb_device **usb_devs = NULL;
 
+struct descriptor_strings {
+    char vendor[64];
+    char product[64];
+    char serial[64];
+    char description[256];
+};
+
 struct hub_info {
     struct libusb_device *dev;
     int bcd_usb;
     int nports;
     int ppps;
     int actionable; /* true if this hub is subject to action */
+    char container_id[33]; /* container ID as hex string */
     char vendor[16];
     char location[32];
-    char description[256];
+    int level;
+    struct descriptor_strings ds;
 };
 
 /* Array of all enumerated USB hubs */
@@ -200,17 +207,19 @@ static int hub_phys_count = 0;
 /* default options */
 static char opt_vendor[16]   = "";
 static char opt_location[32] = "";     /* Hub location a-b.c.d */
+static int opt_level = 0;              /* Hub location level (e.g., a-b is level 2, a-b.c is level 3)*/
 static int opt_ports  = ALL_HUB_PORTS; /* Bitmask of ports to operate on */
 static int opt_action = POWER_KEEP;
-static int opt_delay  = 2;
+static double opt_delay = 2;
 static int opt_repeat = 1;
 static int opt_wait   = 20; /* wait before repeating in ms */
 static int opt_exact  = 0;  /* exact location match - disable USB3 duality handling */
 static int opt_reset  = 0;  /* reset hub after operation(s) */
 
 static const struct option long_options[] = {
-    { "loc",      required_argument, NULL, 'l' },
+    { "location", required_argument, NULL, 'l' },
     { "vendor",   required_argument, NULL, 'n' },
+    { "level",    required_argument, NULL, 'L' },
     { "ports",    required_argument, NULL, 'p' },
     { "action",   required_argument, NULL, 'a' },
     { "delay",    required_argument, NULL, 'd' },
@@ -224,7 +233,7 @@ static const struct option long_options[] = {
 };
 
 
-int print_usage()
+static int print_usage()
 {
     printf(
         "uhubctl %s: utility to control USB port power for smart hubs.\n"
@@ -234,9 +243,10 @@ int print_usage()
         "Options [defaults in brackets]:\n"
         "--action,   -a - action to off/on/cycle (0/1/2) for affected ports.\n"
         "--ports,    -p - ports to operate on    [all hub ports].\n"
-        "--loc,      -l - limit hub by location  [all smart hubs].\n"
+        "--location, -l - limit hub by location  [all smart hubs].\n"
+        "--level     -L - limit hub by location level (e.g. a-b.c is level 3).\n"
         "--vendor,   -n - limit hub by vendor id [%s] (partial ok).\n"
-        "--delay,    -d - delay for cycle action [%d sec].\n"
+        "--delay,    -d - delay for cycle action [%g sec].\n"
         "--repeat,   -r - repeat power off count [%d] (some devices need it to turn off).\n"
         "--exact,    -e - exact location (no USB3 duality handling).\n"
         "--reset,    -R - reset hub after each power-on action, causing all devices to reassociate.\n"
@@ -261,11 +271,83 @@ static char* rtrim(char* str)
 {
     int i;
     for (i = strlen(str)-1; i>=0 && isspace(str[i]); i--) {
-       str[i] = 0;
+        str[i] = 0;
     }
     return str;
 }
 
+/*
+ * Convert port list into bitmap.
+ * Following port list specifications are equivalent:
+ *   1,3,4,5,11,12,13
+ *   1,3-5,11-13
+ * Returns: bitmap of specified ports, max port is MAX_HUB_PORTS.
+ */
+
+static int ports2bitmap(char* const portlist)
+{
+    int ports = 0;
+    char* position = portlist;
+    char* comma;
+    char* dash;
+    int len;
+    int i;
+    while (position) {
+        char buf[8] = {0};
+        comma = strchr(position, ',');
+        len = sizeof(buf) - 1;
+        if (comma) {
+            if (len > comma - position)
+                len = comma - position;
+            strncpy(buf, position, len);
+            position = comma + 1;
+        } else {
+            strncpy(buf, position, len);
+            position = NULL;
+        }
+        /* Check if we have port range, e.g.: a-b */
+        int a=0, b=0;
+        a = atoi(buf);
+        dash = strchr(buf, '-');
+        if (dash) {
+            b = atoi(dash+1);
+        } else {
+            b = a;
+        }
+        if (a > b) {
+            fprintf(stderr, "Bad port spec %d-%d, first port must be less than last\n", a, b);
+            exit(1);
+        }
+        if (a <= 0 || a > MAX_HUB_PORTS || b <= 0 || b > MAX_HUB_PORTS) {
+            fprintf(stderr, "Bad port spec %d-%d, port numbers must be from 1 to %d\n", a, b, MAX_HUB_PORTS);
+            exit(1);
+        }
+        for (i=a; i<=b; i++) {
+            ports |= (1 << (i-1));
+        }
+    }
+    return ports;
+}
+
+
+/*
+ * Compatibility wrapper around libusb_get_port_numbers()
+ */
+
+static int get_port_numbers(libusb_device *dev, uint8_t *buf, uint8_t bufsize)
+{
+    int pcount;
+#if defined(LIBUSB_API_VERSION) && (LIBUSB_API_VERSION >= 0x01000102)
+    /*
+     * libusb_get_port_path is deprecated since libusb v1.0.16,
+     * therefore use libusb_get_port_numbers when supported
+     */
+    pcount = libusb_get_port_numbers(dev, buf, bufsize);
+#else
+    pcount = libusb_get_port_path(NULL, dev, buf, bufsize);
+#endif
+    return pcount;
+}
 
 /*
  * get USB hub properties.
@@ -273,14 +355,13 @@ static char* rtrim(char* str)
  * returns 0 for success and error code for failure.
  */
 
-int get_hub_info(struct libusb_device *dev, struct hub_info *info)
+static int get_hub_info(struct libusb_device *dev, struct hub_info *info)
 {
     int rc = 0;
     int len = 0;
     struct libusb_device_handle *devh = NULL;
     unsigned char buf[LIBUSB_DT_HUB_NONVAR_SIZE + 2 + 3] = {0};
-    struct usb_hub_descriptor *uhd =
-                (struct usb_hub_descriptor *)buf;
+    struct usb_hub_descriptor *uhd = (struct usb_hub_descriptor *)buf;
     int minlen = LIBUSB_DT_HUB_NONVAR_SIZE + 2;
     struct libusb_device_descriptor desc;
     rc = libusb_get_device_descriptor(dev, &desc);
@@ -318,15 +399,8 @@ int get_hub_info(struct libusb_device *dev, struct hub_info *info)
             /* Convert bus and ports array into USB location string */
             int bus = libusb_get_bus_number(dev);
             snprintf(info->location, sizeof(info->location), "%d", bus);
-#if defined(LIBUSB_API_VERSION) && (LIBUSB_API_VERSION >= 0x01000102)
-            /*
-             * libusb_get_port_path is deprecated since libusb v1.0.16,
-             * therefore use libusb_get_port_numbers when supported
-             */
-            int pcount = libusb_get_port_numbers(dev, port_numbers, MAX_HUB_CHAIN);
-#else
-            int pcount = libusb_get_port_path(NULL, dev, port_numbers, MAX_HUB_CHAIN);
-#endif
+            int pcount = get_port_numbers(dev, port_numbers, MAX_HUB_CHAIN);
+            info->level = pcount + 1;
             int k;
             for (k=0; k<pcount; k++) {
                 char s[8];
@@ -349,6 +423,28 @@ int get_hub_info(struct libusb_device *dev, struct hub_info *info)
         } else {
             rc = len;
         }
+        /* Get container_id: */
+        bzero(info->container_id, sizeof(info->container_id));
+        struct libusb_bos_descriptor *bos;
+        rc = libusb_get_bos_descriptor(devh, &bos);
+        if (rc == 0) {
+            int cap;
+            for (cap=0; cap < bos->bNumDeviceCaps; cap++) {
+                if (bos->dev_capability[cap]->bDevCapabilityType == LIBUSB_BT_CONTAINER_ID) {
+                    struct libusb_container_id_descriptor *container_id;
+                    rc = libusb_get_container_id_descriptor(NULL, bos->dev_capability[cap], &container_id);
+                    if (rc == 0) {
+                        int i;
+                        for (i=0; i<16; i++) {
+                            sprintf(info->container_id+i*2, "%02x", container_id->ContainerID[i]);
+                        }
+                        info->container_id[i*2] = 0;
+                        libusb_free_container_id_descriptor(container_id);
+                    }
+                }
+            }
+            libusb_free_bos_descriptor(bos);
+        }
         libusb_close(devh);
     }
     return rc;
@@ -384,9 +480,9 @@ static int get_port_status(struct libusb_device_handle *devh, int port)
 
 
 /*
- * Get USB device description as a string.
+ * Get USB device descriptor strings and summary description.
  *
- * It will use following format:
+ * Summary will use following format:
  *
  *    "<vid:pid> <vendor> <product> <serial>, <USB x.yz, N ports>"
  *
@@ -394,42 +490,40 @@ static int get_port_status(struct libusb_device_handle *devh, int port)
  * may be skipped if they are empty or not enough permissions to read them.
  * <USB x.yz, N ports> will be present only for USB hubs.
  *
- * returns 0 for success and error code for failure.
- * in case of failure description buffer is not altered.
+ * Returns 0 for success and error code for failure.
+ * In case of failure return buffer is not altered.
  */
 
-static int get_device_description(struct libusb_device * dev, char* description, int desc_len)
+static int get_device_description(struct libusb_device * dev, struct descriptor_strings * ds)
 {
     int rc;
     int id_vendor  = 0;
     int id_product = 0;
-    char vendor[64]  = "";
-    char product[64] = "";
-    char serial[64]  = "";
     char ports[64]   = "";
     struct libusb_device_descriptor desc;
     struct libusb_device_handle *devh = NULL;
     rc = libusb_get_device_descriptor(dev, &desc);
     if (rc)
         return rc;
+    bzero(ds, sizeof(*ds));
     id_vendor  = libusb_le16_to_cpu(desc.idVendor);
     id_product = libusb_le16_to_cpu(desc.idProduct);
     rc = libusb_open(dev, &devh);
     if (rc == 0) {
         if (desc.iManufacturer) {
             libusb_get_string_descriptor_ascii(devh,
-                desc.iManufacturer, (unsigned char*)vendor, sizeof(vendor));
-            rtrim(vendor);
+                desc.iManufacturer, (unsigned char*)ds->vendor, sizeof(ds->vendor));
+            rtrim(ds->vendor);
         }
         if (desc.iProduct) {
             libusb_get_string_descriptor_ascii(devh,
-                desc.iProduct, (unsigned char*)product, sizeof(product));
-            rtrim(product);
+                desc.iProduct, (unsigned char*)ds->product, sizeof(ds->product));
+            rtrim(ds->product);
         }
         if (desc.iSerialNumber) {
             libusb_get_string_descriptor_ascii(devh,
-                desc.iSerialNumber, (unsigned char*)serial, sizeof(serial));
-            rtrim(serial);
+                desc.iSerialNumber, (unsigned char*)ds->serial, sizeof(ds->serial));
+            rtrim(ds->serial);
         }
         if (desc.bDeviceClass == LIBUSB_CLASS_HUB) {
             struct hub_info info;
@@ -441,12 +535,12 @@ static int get_device_description(struct libusb_device * dev, char* description,
         }
         libusb_close(devh);
     }
-    snprintf(description, desc_len,
+    snprintf(ds->description, sizeof(ds->description),
         "%04x:%04x%s%s%s%s%s%s%s",
         id_vendor, id_product,
-        vendor[0]  ? " " : "", vendor,
-        product[0] ? " " : "", product,
-        serial[0]  ? " " : "", serial,
+        ds->vendor[0]  ? " " : "", ds->vendor,
+        ds->product[0] ? " " : "", ds->product,
+        ds->serial[0]  ? " " : "", ds->serial,
         ports
     );
     return 0;
@@ -464,9 +558,17 @@ static int print_port_status(struct hub_info * hub, int portmask)
     int port_status;
     struct libusb_device_handle * devh = NULL;
     int rc = 0;
+    int hub_bus;
+    int dev_bus;
+    unsigned char hub_pn[MAX_HUB_CHAIN];
+    unsigned char dev_pn[MAX_HUB_CHAIN];
+    int hub_plen;
+    int dev_plen;
     struct libusb_device *dev = hub->dev;
     rc = libusb_open(dev, &devh);
     if (rc == 0) {
+        hub_bus = libusb_get_bus_number(dev);
+        hub_plen = get_port_numbers(dev, hub_pn, sizeof(hub_pn));
         int port;
         for (port = 1; port <= hub->nports; port++) {
             if (portmask > 0 && (portmask & (1 << (port-1))) == 0) continue;
@@ -481,14 +583,20 @@ static int print_port_status(struct hub_info * hub, int portmask)
 
             printf("  Port %d: %04x", port, port_status);
 
-            char description[256] = "";
+            struct descriptor_strings ds;
+            bzero(&ds, sizeof(ds));
             struct libusb_device * udev;
             int i = 0;
             while ((udev = usb_devs[i++]) != NULL) {
-                if (libusb_get_parent(udev)      == dev &&
+                dev_bus = libusb_get_bus_number(udev);
+                /* only match devices on the same bus: */
+                if (dev_bus != hub_bus) continue;
+                dev_plen = get_port_numbers(udev, dev_pn, sizeof(dev_pn));
+                if ((dev_plen == hub_plen + 1) &&
+                    (memcmp(hub_pn, dev_pn, hub_plen) == 0) &&
                     libusb_get_port_number(udev) == port)
                 {
-                    rc = get_device_description(udev, description, sizeof(description));
+                    rc = get_device_description(udev, &ds);
                     if (rc == 0)
                         break;
                 }
@@ -506,7 +614,7 @@ static int print_port_status(struct hub_info * hub, int portmask)
                     if (port_status & USB_PORT_STAT_SUSPEND)      printf(" suspend");
                 }
             } else {
-                if (port_status == USB_SS_PORT_LS_SS_DISABLED) {
+                if (!(port_status & USB_SS_PORT_STAT_POWER)) {
                     printf(" off");
                 } else {
                     int link_state = port_status & USB_PORT_STAT_LINK_STATE;
@@ -535,7 +643,7 @@ static int print_port_status(struct hub_info * hub, int portmask)
             if (port_status & USB_PORT_STAT_ENABLE)      printf(" enable");
             if (port_status & USB_PORT_STAT_CONNECTION)  printf(" connect");
 
-            if (port_status & USB_PORT_STAT_CONNECTION)  printf(" [%s]", description);
+            if (port_status & USB_PORT_STAT_CONNECTION)  printf(" [%s]", ds.description);
 
             printf("\n");
         }
@@ -573,16 +681,21 @@ static int usb_find_hubs()
         if (rc) {
             perm_ok = 0; /* USB permission issue? */
         }
-        get_device_description(dev, info.description, sizeof(info.description));
+        get_device_description(dev, &info.ds);
         if (info.ppps) { /* PPPS is supported */
             if (hub_count < MAX_HUBS) {
                 info.actionable = 1;
-                if (strlen(opt_location)>0) {
+                if (strlen(opt_location) > 0) {
                     if (strcasecmp(opt_location, info.location)) {
-                       info.actionable = 0;
+                        info.actionable = 0;
                     }
                 }
-                if (strlen(opt_vendor)>0) {
+                if (opt_level > 0) {
+                    if (opt_level != info.level) {
+                        info.actionable = 0;
+                    }
+                }
+                if (strlen(opt_vendor) > 0) {
                     if (strncasecmp(opt_vendor, info.vendor, strlen(opt_vendor))) {
                         info.actionable = 0;
                     }
@@ -592,54 +705,74 @@ static int usb_find_hubs()
             }
         }
     }
+    if (!opt_exact) {
+        /* Handle USB2/3 duality: */
+        for (i=0; i<hub_count; i++) {
+            /* Check only actionable hubs: */
+            if (hubs[i].actionable != 1)
+                continue;
+            /* Must have non empty container ID: */
+            if (strlen(hubs[i].container_id) == 0)
+                continue;
+            int match = -1;
+            for (j=0; j<hub_count; j++) {
+                if (i==j)
+                    continue;
+
+                /* Find hub which is USB2/3 dual to the hub above */
+
+                /* Hub and its dual must be different types: one USB2, another USB3: */
+                if ((hubs[i].bcd_usb < USB_SS_BCD) ==
+                    (hubs[j].bcd_usb < USB_SS_BCD))
+                    continue;
+
+                /* Must have non empty container ID: */
+                if (strlen(hubs[j].container_id) == 0)
+                    continue;
+
+                /* Per USB 3.0 spec chapter 11.2, container IDs must match: */
+                if (strcmp(hubs[i].container_id, hubs[j].container_id) != 0)
+                    continue;
+
+                /* At this point, it should be enough to claim a match.
+                 * However, some devices use hardcoded non-unique container ID.
+                 * We should do few more checks below if multiple such devices are present.
+                 */
+
+                /* If serial numbers are both present, they must match: */
+                if ((strlen(hubs[i].ds.serial) > 0 && strlen(hubs[j].ds.serial) > 0) &&
+                    strcmp(hubs[i].ds.serial, hubs[j].ds.serial) != 0)
+                {
+                    continue;
+                }
+
+                /* Hubs should have the same number of ports: */
+                if (hubs[i].nports != hubs[j].nports)
+                    continue;
+
+                /* And the same level: */
+                if (hubs[i].level != hubs[j].level)
+                    continue;
+
+                /* Finally, we claim a match: */
+                match = j;
+                break;
+            }
+            if (match >= 0) {
+                if (!hubs[match].actionable) {
+                    /* Use 2 to signify that this is derived dual device */
+                    hubs[match].actionable = 2;
+                }
+            }
+        }
+    }
     hub_phys_count = 0;
     for (i=0; i<hub_count; i++) {
-        /* Check only actionable USB3 hubs: */
         if (!hubs[i].actionable)
             continue;
         if (hubs[i].bcd_usb < USB_SS_BCD || opt_exact) {
             hub_phys_count++;
         }
-        if (opt_exact)
-            continue;
-        int match = -1;
-        for (j=0; j<hub_count; j++) {
-            if (i==j)
-                continue;
-
-            /* Find hub which is USB2/3 dual to the hub above.
-             * This is quite reliable and predictable on Linux
-             * but not on Mac, where we may match wrong hub :(
-             * It will work reliably on Mac if there is
-             * only one compatible USB3 hub is connected.
-             * TODO: discover better way to find dual hub.
-             */
-
-            /* Hub and its dual must be different types: one USB2, another USB3: */
-            if ((hubs[i].bcd_usb < USB_SS_BCD) ==
-                (hubs[j].bcd_usb < USB_SS_BCD))
-                continue;
-
-            /* But they must have the same vendor: */
-            if (strncasecmp(hubs[i].vendor, hubs[j].vendor, 4))
-                continue;
-
-            /* Provisionally we choose this one as dual: */
-            if (match < 0 && !hubs[j].actionable)
-                match = j;
-
-            /* But if there is exact port path match,
-             * we prefer it (true for Linux but not Mac):
-             */
-            char *p1 = strchr(hubs[i].location, '-');
-            char *p2 = strchr(hubs[j].location, '-');
-            if (p1 && p2 && strcasecmp(p1, p2)==0) {
-                match = j;
-                break;
-            }
-        }
-        if (match >= 0)
-            hubs[match].actionable = 1;
     }
     if (perm_ok == 0 && hub_phys_count == 0) {
         return LIBUSB_ERROR_ACCESS;
@@ -655,7 +788,7 @@ int main(int argc, char *argv[])
     int option_index = 0;
 
     for (;;) {
-        c = getopt_long(argc, argv, "l:n:a:p:d:r:w:hveR",
+        c = getopt_long(argc, argv, "l:L:n:a:p:d:r:w:hveR",
             long_options, &option_index);
         if (c == -1)
             break;  /* no more options left */
@@ -672,6 +805,9 @@ int main(int argc, char *argv[])
         case 'l':
             strncpy(opt_location, optarg, sizeof(opt_location));
             break;
+        case 'L':
+            opt_level = atoi(optarg);
+            break;
         case 'n':
             strncpy(opt_vendor, optarg, sizeof(opt_vendor));
             break;
@@ -680,16 +816,7 @@ int main(int argc, char *argv[])
                 break;
             }
             if (strlen(optarg)) {
-                /* parse port list */
-                opt_ports = 0;
-                size_t i;
-                for (i=0; i<strlen(optarg); i++) {
-                    if (!isdigit(optarg[i]) || optarg[i] == '0') {
-                        printf("%s must be list of ports 1 to %d\n", optarg, MAX_HUB_PORTS);
-                    }
-                    int d = optarg[i]-'1';
-                    opt_ports |= (1 << d);
-                }
+                opt_ports = ports2bitmap(optarg);
             }
             break;
         case 'a':
@@ -704,7 +831,7 @@ int main(int argc, char *argv[])
             }
             break;
         case 'd':
-            opt_delay = atoi(optarg);
+            opt_delay = atof(optarg);
             break;
         case 'r':
             opt_repeat = atoi(optarg);
@@ -790,28 +917,30 @@ int main(int argc, char *argv[])
         );
         exit(1);
     }
-    int i;
-    for (i=0; i<hub_count; i++) {
-        if (hubs[i].actionable == 0)
+    int k; /* k=0 for power OFF, k=1 for power ON */
+    for (k=0; k<2; k++) { /* up to 2 power actions - off/on */
+        if (k == 0 && opt_action == POWER_ON )
             continue;
-        printf("Current status for hub %s [%s]\n",
-            hubs[i].location, hubs[i].description
-        );
-        print_port_status(&hubs[i], opt_ports);
-        if (opt_action == POWER_KEEP) { /* no action, show status */
+        if (k == 1 && opt_action == POWER_OFF)
             continue;
-        }
-        struct libusb_device_handle * devh = NULL;
-        rc = libusb_open(hubs[i].dev, &devh);
-        if (rc == 0) {
-            /* will operate on these ports */
-            int ports = ((1 << hubs[i].nports) - 1) & opt_ports;
-            int k; /* k=0 for power OFF, k=1 for power ON */
-            for (k=0; k<2; k++) { /* up to 2 power actions - off/on */
-                if (k == 0 && opt_action == POWER_ON )
-                    continue;
-                if (k == 1 && opt_action == POWER_OFF)
-                    continue;
+        if (k == 1 && opt_action == POWER_KEEP)
+            continue;
+        int i;
+        for (i=0; i<hub_count; i++) {
+            if (hubs[i].actionable == 0)
+                continue;
+            printf("Current status for hub %s [%s]\n",
+                hubs[i].location, hubs[i].ds.description
+            );
+            print_port_status(&hubs[i], opt_ports);
+            if (opt_action == POWER_KEEP) { /* no action, show status */
+                continue;
+            }
+            struct libusb_device_handle * devh = NULL;
+            rc = libusb_open(hubs[i].dev, &devh);
+            if (rc == 0) {
+                /* will operate on these ports */
+                int ports = ((1 << hubs[i].nports) - 1) & opt_ports;
                 int request = (k == 0) ? LIBUSB_REQUEST_CLEAR_FEATURE
                                        : LIBUSB_REQUEST_SET_FEATURE;
                 int port;
@@ -819,7 +948,7 @@ int main(int argc, char *argv[])
                     if ((1 << (port-1)) & ports) {
                         int port_status = get_port_status(devh, port);
                         int power_mask = hubs[i].bcd_usb < USB_SS_BCD ? USB_PORT_STAT_POWER
-                                                                          : USB_SS_PORT_STAT_POWER;
+                                                                      : USB_SS_PORT_STAT_POWER;
                         if (k == 0 && !(port_status & power_mask))
                             continue;
                         if (k == 1 && (port_status & power_mask))
@@ -844,8 +973,6 @@ int main(int argc, char *argv[])
                         }
                     }
                 }
-                if (k==0 && opt_action == POWER_CYCLE)
-                    sleep_ms(opt_delay * 1000);
                 /* USB3 hubs need extra delay to actually turn off: */
                 if (k==0 && hubs[i].bcd_usb >= USB_SS_BCD)
                     sleep_ms(150);
@@ -853,7 +980,7 @@ int main(int argc, char *argv[])
                     request == LIBUSB_REQUEST_CLEAR_FEATURE ? "off" : "on"
                 );
                 printf("New status for hub %s [%s]\n",
-                    hubs[i].location, hubs[i].description
+                    hubs[i].location, hubs[i].ds.description
                 );
                 print_port_status(&hubs[i], opt_ports);
 
@@ -869,6 +996,8 @@ int main(int argc, char *argv[])
             }
             libusb_close(devh);
         }
+        if (k == 0 && opt_action == POWER_CYCLE)
+            sleep_ms((int)(opt_delay * 1000));
     }
     rc = 0;
 cleanup:
